package edu.utexas.cs.tamerProject.applet;

/*
By Brad Knox.
*/
import java.applet.AppletContext;
import java.awt.Color;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.util.*;

import java.util.Properties;


import javax.swing.JApplet;

import org.rlcommunity.rlglue.codec.AgentInterface;
import org.rlcommunity.rlglue.codec.EnvironmentInterface;


public class RLApplet extends JApplet 
			implements Observer {
	private static final long serialVersionUID = 69786163333408557L;
	
	protected RLPanel rlPanel = null;
	protected boolean hasBeenReset = false;
	private final int STEPS_PER_LOG_SAVE = 30; // saves at episode end too
	
	public void init() {
		RunLocalExperiment.isApplet = true;
		int height = this.getHeight();
		int width = this.getWidth();
		if (rlPanel == null) {
			rlPanel = makeRLPanel();
			rlPanel.setSize(width, height);
		}
		
		prepForInitPanel();
		
		initPanel(null, null);
		
		System.out.println("\n\n\nEnd of init()\n\n\n");
	}
	
	public RLPanel makeRLPanel() {
		return new RLPanel();
	}
	
	
	public void prepForInitPanel() {
		this.setBackground(Color.black);

		// Get max number of episodes from parameters set in HTML; will exit once this limit is reached
		String numEpsStr = getParameter("numEpisodes");
		if (numEpsStr != null) RunLocalExperiment.numEpisodes = Integer.parseInt(numEpsStr);

		// Get max total number of steps from parameters set in HTML; will exit once this limit is reached
		String maxTotalStepsStr = getParameter("maxTotalSteps");
		if (maxTotalStepsStr != null)
			RunLocalExperiment.maxTotalSteps = Integer.parseInt(maxTotalStepsStr);
			
		// If set, this gives a number of time steps that one episode must last before the experiment will
		// end. This can be used while training a subject before an experiment to force a certain control
		// or training performance before moving to the real experiment.
		String finishExpAtStepsStr = getParameter("timeRequirement");
		if (finishExpAtStepsStr != null)
			RunLocalExperiment.finishExpIfNumStepsInOneEp = Integer.parseInt(finishExpAtStepsStr);		
				
		// Get the duration of a time step in milliseconds. Smaller values create faster transitions.
		String speedStr = getParameter("speed");
		System.out.println("speed: " + speedStr);
		if (speedStr != null) 
			RunLocalExperiment.stepDurInMilliSecs = Integer.valueOf(speedStr);
		
		// Boolean value determines whether the keys '+' and '-' can speed and slow transitions.
		String speedControlsStr = getParameter("speedControls");
		if (speedControlsStr != null) 
			RLPanel.enableSpeedControls = Boolean.valueOf(speedControlsStr);

		// Boolean value determines whether the key '1' can be used to move one step forward (only useful when paused).
		String singleStepControlStr = getParameter("singleStepControl");
		if (singleStepControlStr != null) 
			RLPanel.enableSingleStepControl = Boolean.valueOf(singleStepControlStr);
		
	}
	
	
	public void initPanel(AgentInterface agent, EnvironmentInterface env) {
		rlPanel.init(agent, env);
		this.getContentPane().add(rlPanel);
		this.rlPanel.runLocal.addObserver(this);
		rlPanel.runLocal.initExp();
		rlPanel.runLocal.startExp();
	}
	
    /*
     * Called when TinyState object (in RunLocal) changes 
     */
	public void update(Observable observable, Object obj) {
		boolean timeToLog = false;
		if (this.rlPanel.runLocal.glue.getTimeStep() % STEPS_PER_LOG_SAVE == 0) // i.e., save every n steps and at the end of the episode
			timeToLog = true;
		if (observable instanceof RunLocalExperiment && obj instanceof Boolean) {
			boolean endOfEp = ((Boolean)obj).booleanValue();
			if (endOfEp)
				timeToLog = true;
		}
		if (timeToLog) {
			saveLog();
		}
	}
	
	public void reset() {
		this.requestFocusInWindow();
		this.getContentPane().removeAll();
		this.hasBeenReset = true;
		this.init();
	}
	
	

	
	public void saveLog() {}
	

}
