package edu.utexas.cs.tamerProject.applet;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.util.Date;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.Timer;

import org.rlcommunity.rlglue.codec.AgentInterface;
import org.rlcommunity.rlglue.codec.EnvironmentInterface;
import org.rlcommunity.rlviz.app.VisualizerFactory;
import org.rlcommunity.rlviz.app.VisualizerPanel;
import org.rlcommunity.rlviz.app.frames.VisualizerVizFrame;

import rlVizLib.messaging.environment.EnvVisualizerNameRequest;
import rlVizLib.messaging.environment.EnvVisualizerNameResponse;
import rlVizLib.visualization.AbstractVisualizer;
import rlVizLib.visualization.RenderObject;
import rlVizLib.visualization.VisualizerPanelInterface;

public class RLPanel extends JPanel 
					implements Observer, KeyListener, MouseListener, VisualizerPanelInterface {
	private static final long serialVersionUID = 111112222333445L;
	
	RunLocalExperiment runLocal;
	AgentInterface agent;
	EnvironmentInterface env;
	
	Timer repaintTimer;
	AbstractVisualizer theEnvViz;
	public static String defaultEnvVisualizerClassName = 
					"org.rlcommunity.visualizers.generic.GenericEnvVisualizer";
	BufferedImage bufferedEnvImage = null;
	protected final Object envImageLock = new Object();
	
	final double DUR_OF_DISPLAY = 3; // in seconds
	final String nameForTimeSteps = "Time"; //"Time step"
	final boolean DISPLAY_SECONDS_FOR_TIME = false; // display time step number if false. TODO return to true for turk
	final String nameForEpisodes = "Game number"; //"Episode"
	double timeOfLastDurChange = getCurrentTimeInSecs();
	double timeOfLastEpStart = getCurrentTimeInSecs();
	
	public static boolean enableSpeedControls = true;
	public static boolean enableSingleStepControl = true;
		
	
	public void init(AgentInterface agent, EnvironmentInterface env) {	
		setFocusable(true);
		setFocusTraversalKeysEnabled(false);
		setRequestFocusEnabled(true);
		addKeyListener(this);
		addMouseListener(this);
		setBackground(Color.white);	
		
		runLocal = new RunLocalExperiment();
		runLocal.theAgent = agent;
		runLocal.theEnvironment = env;
		runLocal.init();
		this.agent = runLocal.theAgent;
		this.env = runLocal.theEnvironment;
        EnvVisualizerNameResponse theNameResponse = EnvVisualizerNameRequest.Execute();
        String vizClassName = theNameResponse.getTheVisualizerClassName();
        theEnvViz = createVisualizer(vizClassName);
        theEnvViz.setParentPanel(this);

		theEnvViz.startVisualizing();
		theEnvViz.notifyPanelSizeChange();
		
	    ActionListener actionListener = new ActionListener() {
	    	public void actionPerformed(ActionEvent actionEvent) {
	    		if (runLocal.glue.getTimeStep() == 1) // This is a hack to make show the initial state in case the agent takes a while to respond with an action.
	    			update(null, null);
	    		repaint();
	        }
	    };
	    repaintTimer = new Timer(33, actionListener);
	    repaintTimer.start();
		requestFocus();
	}
	
	
	

	public void paint (Graphics g) {
    	paint(g, getCurrentTimeInSecs());
	}
    public void paint( Graphics g, double currTime ) {
    	if (!runLocal.expInitialized)
    		return;
    	synchronized(envImageLock) {
    		super.paint(g);    		
    		g.drawImage(bufferedEnvImage, 0, -12, this.getWidth(), this.getHeight(), this);
    	}
    	requestFocusInWindow();
    }
    

    protected void drawMeta(Graphics2D g, double currTime) {
    	Color startColor = g.getColor();
    	g.setColor(Color.black);
    	if (RLPanel.enableSpeedControls && 
    			currTime - this.timeOfLastDurChange < DUR_OF_DISPLAY) {
    		String speedStr = "Steps per second: " + 1000.0 / ((int)RunLocalExperiment.stepDurInMilliSecs);
    		g.drawString(speedStr, 
    				4*g.getFontMetrics().charWidth(' '), 
    				g.getFont().getSize() + 12);
    	}
    	
    	// draw time on panel
    	int stepNum = runLocal.glue.getTimeStep();
    	int time = DISPLAY_SECONDS_FOR_TIME ? (int)(Math.round((float)stepNum 
    						* RunLocalExperiment.stepDurInMilliSecs * 0.001)) 
    						: stepNum;
    	String timeStr = nameForTimeSteps + ": " + time;
    	//System.out.println("timeStr: " + timeStr);
    	g.drawString(timeStr,
    			(this.getWidth() - g.getFontMetrics().stringWidth(timeStr) - 4*g.getFontMetrics().charWidth(' ')), 
    			this.getHeight() - (int)(0.2 * g.getFontMetrics().getHeight()));
    	
    	// draw episode number if a new episode recently began
    	if (currTime - this.timeOfLastEpStart < DUR_OF_DISPLAY) {
    		String epStr = nameForEpisodes + ": " + runLocal.glue.getEpisodeNumber(); 
    		g.drawString(epStr,
        			(this.getWidth() - g.getFontMetrics().stringWidth(epStr)) / 2, 
        			this.getHeight() - (int)(0.2 * g.getFontMetrics().getHeight()));
    	}
    	g.setColor(startColor);
    }
    
    /*
     * Called when any observed Observable object changes. This class observes
     * runLocal and the TinyState object in runLocal. 
     */
	public void update(Observable observable, Object obj) {
	}
	
	private void makeCurrentEnvImage() {
		synchronized(envImageLock) {
			this.bufferedEnvImage = getImageCopy(theEnvViz.getLatestImage());
			drawMeta(bufferedEnvImage.createGraphics(), getCurrentTimeInSecs());
		}
	}


	
	
	
	
	public void keyPressed( KeyEvent e ) {
		System.out.println("pressed in rlpanel: " + e.getKeyChar());
		boolean stepDurChange = false;
		if (RLPanel.enableSpeedControls 
				&& e.getKeyChar() == '+') { // speed up
			stepDurChange = true;
			RunLocalExperiment.stepDurInMilliSecs = Math.max(RunLocalExperiment.stepDurInMilliSecs / Math.cbrt(2.0), 1); // speed doubles every 3 presses
		}
		else if (RLPanel.enableSpeedControls 
				&& e.getKeyChar() == '-') { // slow down
			stepDurChange = true;
			RunLocalExperiment.stepDurInMilliSecs = Math.max((RunLocalExperiment.stepDurInMilliSecs * Math.cbrt(2.0)), 1.0); // speed halves every 3 presses
		}
		else if (e.getKeyChar() == '0') { // stop exp
			runLocal.stopExp();
		}
		else if (RLPanel.enableSingleStepControl 
				&& e.getKeyChar() == '1') { // take step (only makes sense when stopped)
			runLocal.takeStep();
		}
		else if (e.getKeyChar() == '2') { // start exp
			runLocal.startExp();
		}
		if (stepDurChange) { // handle change in step duration
			runLocal.stopExp();
			timeOfLastDurChange = getCurrentTimeInSecs();
			runLocal.startExp();
		}
		if (e.getKeyChar() == 'u') { // start exp
			update(null, null);
		}
		e.consume();
		
	}
	public void keyReleased( KeyEvent e ) {}
	public void keyTyped( KeyEvent e ) {}

	
	public void mouseEntered( MouseEvent e ) { }
	public void mouseExited( MouseEvent e ) { }
	public void mousePressed( MouseEvent e ) { }
	public void mouseReleased( MouseEvent e ) { }
	public void mouseClicked( MouseEvent e ) {
		requestFocus();
	}
	
	
	
	
	
	public static double getCurrentTimeInSecs(){
		return (new Date()).getTime() / 1000.0;
	}
	
	protected BufferedImage getImageCopy(BufferedImage orig) {
	    BufferedImage copy = new BufferedImage(this.getWidth(), this.getHeight(),
	        orig.getType());
	    Graphics g = copy.createGraphics();
	    g.drawImage(orig, (this.getWidth() - orig.getWidth()) / 2, 
	    		this.getHeight() - orig.getHeight(), null);
	    g.dispose();
	    return copy;
	}
	
	
	// this method is adapted from Brian Tanner's VisualizerFactory class
	public AbstractVisualizer createVisualizer(String vizClassName) {
		Class<?> theClass = null;
		try {
			theClass = Class.forName(vizClassName);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
		AbstractVisualizer theViz = null;
	    //First try and load the one with all parameters
	    theViz = VisualizerFactory.createVisualizer(theClass, runLocal.glue, null);
	    if (theViz == null) { //IF that didn't work, try without theControlTarget
	        theViz = VisualizerFactory.createVisualizer(theClass, runLocal.glue);
	    }
	    if (theViz == null) { //IF that didn't work, try without theGlueState
	        theViz = VisualizerFactory.createVisualizer(theClass);
	    }
	    return theViz;
	}





	public void receiveNotificationVizChanged() {
		
		// create the environmental image to be displayed
		if (SwingUtilities.isEventDispatchThread()) {
			makeCurrentEnvImage();
		}
		else { // do image update in event dispatch thread
			try {
				SwingUtilities.invokeAndWait(new Runnable() {
				    public void run() {makeCurrentEnvImage();}
				});
			} catch (Exception e) {e.printStackTrace();}
		}
		if (runLocal.glue.getTimeStep() == 1)
			this.timeOfLastEpStart = getCurrentTimeInSecs();
		
	}
			
}


